package sdanderson.androidcharts.BarChart;

import android.content.Context;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.RectF;
import android.graphics.Region;
import android.util.AttributeSet;
import android.view.View;

import java.util.ArrayList;
import java.util.List;

import sdanderson.androidcharts.Common.CoreSettings;
import sdanderson.androidcharts.Common.NumberUtilities;

/**
 * Created by Shane Anderson on 10/4/2015.
 *
 * This is the view of the actual bar chart plotting area
 *
 */
class BarChartInnerView extends View {

    private float mBorderWidth = 6f;
    private int mBorderColor = Color.BLACK;
    private int mBackgroundColor = Color.WHITE;
    private int mGridlineColor = Color.GRAY;
    private float mGridlineWidth = 1f;
    private int mNumOfGridlines = 5;
    private int mZeroLineColor = Color.BLACK;
    private float mZeroLineWidth = 4f;
    private boolean mUseAnimation = false;
    private float mPadding = mBorderWidth / 2f;
    private boolean mRoundOutChartYMinMax = false;
    private int mRoundOutChartYMinMaxHighestDigits = 1;
    private float mHorizontalPadding = 0.05f;
    private float mBarWidths = 0f;

    private List<BarChartItem> mBars;
    private List<Region> mTouchRegions;

    private RectF mRectF = new RectF();
    private Paint mDrawingPaint = new Paint();
    private RectF mDrawingRect = new RectF();

    // Variables to support animations
    private boolean mFinishedDrawing = false;
    private float mAnimationTime = 500f; // in milliseconds
    private long mAnimationElapsedTime = 0;
    private long mAnimationStartTime = 0;
    private boolean mAnimationCounterStarted = false;
    private float mCurrentPercentToDraw = 0.0f;
    private Runnable mRedrawRunnable = new Runnable() {
        public void run() {
            try {
                if (!mFinishedDrawing) {
                    if (!mAnimationCounterStarted) {
                        mAnimationStartTime = System.currentTimeMillis();
                        mAnimationCounterStarted = true;
                    }
                    mAnimationElapsedTime = System.currentTimeMillis() - mAnimationStartTime;
                    invalidate();
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    };

    public BarChartInnerView(Context pContext) {
        super(pContext);
        init();
    }

    public BarChartInnerView(Context pContext, AttributeSet pAttrs) {
        super(pContext, pAttrs);
        init();
    }

    public BarChartInnerView(Context pContext, AttributeSet pAttrs, int pDefStyle) {
        super(pContext, pAttrs, pDefStyle);
        init();
    }

    private void init() {
        mBars = new ArrayList<>();
        setWillNotDraw(false);
        this.setBackgroundColor(Color.TRANSPARENT);
    }

    protected boolean areThereBars() {
        return (mBars != null && mBars.size() > 0);
    }

    protected void addBar (String pLabel, float pValue, int pColor) {
        BarChartItem newBarChartItem = new BarChartItem(pLabel, pValue, pColor);
        mBars.add(newBarChartItem);

        if (mTouchRegions == null)
            mTouchRegions = new ArrayList<>();

        mTouchRegions.add(new Region());
    }

    @SuppressWarnings("unused")
    protected void removeBarByColor (int pColor) {

        if (mBars != null && !mBars.isEmpty()) {
            int index = -1;

            for (int i = 0; i < mBars.size(); i++) {
                if (mBars.get(i).getColor() == pColor)
                    index = i;
            }

            if (index != -1)
                mBars.remove(index);
        }
    }

    @SuppressWarnings("unused")
    protected void removeBarByIndex (int pIndex) {
        if (mBars != null && !mBars.isEmpty() && pIndex < mBars.size()) {
            mBars.remove(pIndex);
        }
    }

    protected BarChartItem getBarChartItemByColor (int pColor) {
        if (mBars != null && !mBars.isEmpty()) {
            for (BarChartItem item : mBars) {
                if (item.getColor() == pColor)
                    return item;
            }
            return null;
        } else {
            return null;
        }
    }

    @SuppressWarnings("unused")
    protected BarChartItem getBarChartItemByIndex (int pIndex) {
        if (mBars != null && !mBars.isEmpty() && pIndex < mBars.size()) {
            return mBars.get(pIndex);
        } else {
            return null;
        }
    }

    @SuppressWarnings("unused")
    protected Integer getBarChartItemIndexFromColor(int pColor) {
        Integer index = null;
        if (mBars != null && !mBars.isEmpty()) {
            for (int i = 0; i < mBars.size(); i++) {
                if (mBars.get(i).getColor() == pColor)
                    index = i;
            }
        }
        return index;
    }

    protected Float getLargestBarValueChartable () {
        if (mBars.size() > 0) {
            Float temp = mBars.get(0).getValue();

            for (BarChartItem item: mBars) {
                if (item.getValue() > temp)
                    temp = item.getValue();
            }

            if (mRoundOutChartYMinMax)
                return NumberUtilities.roundToNearestHighestPlaces(temp,
                                    mRoundOutChartYMinMaxHighestDigits, true).floatValue();
            else
                return temp;
        } else
            return null;
    }

    protected Float getSmallestBarValueChartable () {
        if (mBars.size() > 0) {
            Float temp = mBars.get(0).getValue();

            for (BarChartItem item: mBars) {
                if (item.getValue() < temp)
                    temp = item.getValue();
            }

            if (mRoundOutChartYMinMax)
                return NumberUtilities.roundToNearestHighestPlaces(temp,
                                    mRoundOutChartYMinMaxHighestDigits, false).floatValue();
            else
                return temp;
        } else
            return null;
    }

    @SuppressWarnings("unused")
    protected String getBarLabel(int pBarIndex) {
        if (mBars.size() > pBarIndex)
            return mBars.get(pBarIndex).getLabel();
        else
            return null;
    }

    @SuppressWarnings("unused")
    protected Float getBarValue(int pBarIndex) {
        if (mBars.size() > pBarIndex)
            return mBars.get(pBarIndex).getValue();
        else
            return null;
    }

    @SuppressWarnings("unused")
    protected Integer getBarColor(int pBarIndex) {
        if (mBars.size() > pBarIndex)
            return mBars.get(pBarIndex).getColor();
        else
            return null;
    }

    protected void setAllBarBorderWidths(float pWidth) {
        if (mBars != null && !mBars.isEmpty()) {
            for (BarChartItem item : mBars) {
                item.setBarBorderWidth(pWidth);
            }
        }
    }

    @SuppressWarnings("unused")
    protected void setBarBorderWidthByIndex(float pWidth, int pBarIndex) {
        if (mBars != null && !mBars.isEmpty() && mBars.size() > pBarIndex)
            mBars.get(pBarIndex).setBarBorderWidth(pWidth);
    }

    protected void setAllBarBorderColors(int pColor) {
        if (mBars != null && !mBars.isEmpty()) {
            for (BarChartItem item : mBars) {
                item.setBarBorderColor(pColor);
            }
        }
    }

    @SuppressWarnings("unused")
    protected void setBarBorderColorByIndex(int pColor, int pBarIndex) {
        if (mBars != null && !mBars.isEmpty() && mBars.size() > pBarIndex)
            mBars.get(pBarIndex).setBarBorderColor(pColor);
    }

    protected void setChartBackgroundColor (int pColor) { mBackgroundColor = pColor; }

    protected void setGridlineWidth (float pGridlineWidth) { mGridlineWidth = pGridlineWidth; }

    // This is to set the number of gridlines on both sides of the Y=0 line
    protected void setNumOfGridlines (int pNumOfGridlines) { mNumOfGridlines = pNumOfGridlines; }

    protected void setGridlineColor (int pColor) { mGridlineColor = pColor; }

    protected void setRoundOutChartYMinMax (boolean pRound, int pHighestDigits) {
        mRoundOutChartYMinMax = pRound;
        mRoundOutChartYMinMaxHighestDigits = pHighestDigits;
    }

    protected void setZeroLineWidth (float pWidth) { mZeroLineWidth = pWidth; }

    protected void setZeroLineColor (int pColor) { mZeroLineColor = pColor; }

    protected void setBorderColor(int pColor) { mBorderColor = pColor; }

    protected void setBorderWidth(float pWidth) { mBorderWidth = pWidth; }

    protected void setUseAnimation(boolean pUseAnimation) { mUseAnimation = pUseAnimation; }

    protected void setAnimationTime(float pAnimationTime) { mAnimationTime = pAnimationTime;}

    private void calculateBarWidths() {
        mBarWidths = 0f;
        if (mBars != null && mBars.size() > 0) {
            if (mRectF != null) {
                Float spaceForBars = mRectF.width() -
                                                (mBars.size()+1)*mHorizontalPadding*mRectF.width();
                mBarWidths = spaceForBars/mBars.size();
            }
        }
    }

    protected Float actualYToDrawingY(float pActualY) {
        Float actualY = pActualY;

        Float largestBarValueChartable = getLargestBarValueChartable();
        Float smallestBarValueChartable = getSmallestBarValueChartable();
        if (largestBarValueChartable.floatValue() == smallestBarValueChartable.floatValue()) {
            if (smallestBarValueChartable > 0f) {
                smallestBarValueChartable = 0f;
            } else {
                largestBarValueChartable = 0f;
            }
        }

        // To perform the conversion, shift everything up by the smallest value if that is negative
        if (smallestBarValueChartable < 0f) {
            largestBarValueChartable -= smallestBarValueChartable;
            actualY = pActualY - smallestBarValueChartable;
            smallestBarValueChartable = 0f;
        }

        return ((1-(actualY/(largestBarValueChartable-smallestBarValueChartable)))*
                                            (mRectF.bottom-mRectF.top))+mRectF.top;
    }

    protected Float getYAxisSmallestValue() {
        return getSmallestBarValueChartable();
    }

    protected Float getYAxisLargestValue() {
        return getLargestBarValueChartable();
    }

    @Override
    public void onDraw(final Canvas pCanvas) {
        if (pCanvas == null)
            return;

        mPadding = mBorderWidth / 2f;
        pCanvas.drawColor(mBackgroundColor);

        if (mRectF != null && mDrawingPaint != null && !isInEditMode()) {
            mRectF.set(mPadding, mPadding, getWidth() - mPadding, getHeight() - mPadding);

            // If there is a mix of negative and positive values
            if (getLargestBarValueChartable() > 0f && getSmallestBarValueChartable() < 0f) {
                // Draw horizontal grid lines on both side of Y=0
                // First draw horizontal grid lines for above Y=0
                Float spacing = (actualYToDrawingY(0f)-mRectF.top) / (mNumOfGridlines+1);
                for (int i = 1; i <= mNumOfGridlines; i++) {
                    mDrawingPaint.setColor(mGridlineColor);
                    mDrawingPaint.setStrokeWidth(mGridlineWidth);
                    mDrawingPaint.setStyle(Paint.Style.STROKE);
                    mDrawingPaint.setAntiAlias(true);
                    pCanvas.drawLine(mRectF.left, mRectF.top+(spacing*i), mRectF.right,
                            mRectF.top+(spacing*i), mDrawingPaint);
                }

                // Now draw horizontal grid lines for below Y=0
                spacing = (mRectF.bottom-actualYToDrawingY(0f)) / (mNumOfGridlines+1);
                for (int i = 1; i <= mNumOfGridlines; i++) {
                    mDrawingPaint.setColor(mGridlineColor);
                    mDrawingPaint.setStrokeWidth(mGridlineWidth);
                    mDrawingPaint.setStyle(Paint.Style.STROKE);
                    mDrawingPaint.setAntiAlias(true);
                    pCanvas.drawLine(mRectF.left, mRectF.bottom-(spacing*i), mRectF.right,
                            mRectF.bottom-(spacing*i), mDrawingPaint);
                }
            } else {
                // In this case all values are either positive or all are negative, so just space
                // gridlines out evenly from the max to min
                Float spacing = mRectF.height() / (mNumOfGridlines+1);
                for (int i = 1; i <= mNumOfGridlines; i++) {
                    mDrawingPaint.setColor(mGridlineColor);
                    mDrawingPaint.setStrokeWidth(mGridlineWidth);
                    mDrawingPaint.setStyle(Paint.Style.STROKE);
                    mDrawingPaint.setAntiAlias(true);
                    pCanvas.drawLine(mRectF.left, mRectF.top+(spacing*i), mRectF.right,
                            mRectF.top+(spacing*i), mDrawingPaint);
                }
            }

            // The Titles and Axis text is handled by the BarChart ViewGroup

            calculateBarWidths();

            if (!mUseAnimation) {
                // Draw the bars without animations
                if (mBars != null && !mBars.isEmpty()) {
                    for (int i = 0; i < mBars.size(); i++) {
                        mDrawingPaint.setColor(mBars.get(i).getColor());
                        mDrawingPaint.setStyle(Paint.Style.FILL);
                        mDrawingPaint.setAntiAlias(true);

                        Float currentX = (mRectF.width()-mPadding*2)*mHorizontalPadding +
                                i*(mBarWidths+(mRectF.width()-mPadding*2)*mHorizontalPadding);

                        Float tempY = mBars.get(i).getValue();
                        if (tempY > 0f) {
                            pCanvas.drawRect(currentX, actualYToDrawingY(mBars.get(i).getValue()),
                                    currentX + mBarWidths, actualYToDrawingY(0f), mDrawingPaint);
                            mDrawingPaint.setColor(mBars.get(i).getBarBorderColor());
                            mDrawingPaint.setStrokeWidth(mBars.get(i).getBarBorderWidth());
                            mDrawingPaint.setStyle(Paint.Style.STROKE);
                            pCanvas.drawRect(currentX, actualYToDrawingY(mBars.get(i).getValue()),
                                    currentX + mBarWidths, actualYToDrawingY(0f), mDrawingPaint);
                            // Generate the touch regions
                            mTouchRegions.get(i).set(currentX.intValue(),
                                    actualYToDrawingY(mBars.get(i).getValue()).intValue(),
                                    (int)(currentX + mBarWidths), actualYToDrawingY(0f).intValue());
                        }
                        else {
                            pCanvas.drawRect(currentX, actualYToDrawingY(0f), currentX + mBarWidths,
                                    actualYToDrawingY(mBars.get(i).getValue()), mDrawingPaint);
                            mDrawingPaint.setColor(mBars.get(i).getBarBorderColor());
                            mDrawingPaint.setStrokeWidth(mBars.get(i).getBarBorderWidth());
                            mDrawingPaint.setStyle(Paint.Style.STROKE);
                            pCanvas.drawRect(currentX, actualYToDrawingY(0f), currentX + mBarWidths,
                                    actualYToDrawingY(mBars.get(i).getValue()), mDrawingPaint);
                            // Generate the touch regions
                            mTouchRegions.get(i).set(currentX.intValue(),
                                    actualYToDrawingY(0f).intValue(),
                                    (int)(currentX + mBarWidths),
                                    actualYToDrawingY(mBars.get(i).getValue()).intValue());
                        }
                    }
                }
            } else {
                // Draw the lines using animations

                if (mBars != null && !mBars.isEmpty()) {
                    for (int i = 0; i < mBars.size(); i++) {
                        mDrawingPaint.setColor(mBars.get(i).getColor());
                        mDrawingPaint.setStyle(Paint.Style.FILL);
                        mDrawingPaint.setAntiAlias(true);

                        Float currentX = (mRectF.width()-mPadding*2)*mHorizontalPadding +
                                i*(mBarWidths+(mRectF.width()-mPadding*2)*mHorizontalPadding);

                        Float tempY = mBars.get(i).getValue()*(mCurrentPercentToDraw);
                        if (tempY > 0f) {
                            pCanvas.drawRect(currentX, actualYToDrawingY(tempY),
                                    currentX + mBarWidths, actualYToDrawingY(0f), mDrawingPaint);
                            mDrawingPaint.setColor(mBars.get(i).getBarBorderColor());
                            mDrawingPaint.setStrokeWidth(mBars.get(i).getBarBorderWidth());
                            mDrawingPaint.setStyle(Paint.Style.STROKE);
                            pCanvas.drawRect(currentX, actualYToDrawingY(tempY),
                                    currentX + mBarWidths, actualYToDrawingY(0f), mDrawingPaint);
                            // Generate the touch regions
                            mTouchRegions.get(i).set(currentX.intValue(),
                                    actualYToDrawingY(tempY).intValue(),
                                    (int)(currentX + mBarWidths), actualYToDrawingY(0f).intValue());
                        }
                        else {
                            pCanvas.drawRect(currentX, actualYToDrawingY(0f), currentX + mBarWidths,
                                    actualYToDrawingY(tempY), mDrawingPaint);
                            mDrawingPaint.setColor(mBars.get(i).getBarBorderColor());
                            mDrawingPaint.setStrokeWidth(mBars.get(i).getBarBorderWidth());
                            mDrawingPaint.setStyle(Paint.Style.STROKE);
                            pCanvas.drawRect(currentX, actualYToDrawingY(0f), currentX + mBarWidths,
                                    actualYToDrawingY(tempY), mDrawingPaint);
                            // Generate the touch regions
                            mTouchRegions.get(i).set(currentX.intValue(),
                                    actualYToDrawingY(0f).intValue(),
                                    (int)(currentX + mBarWidths),
                                    actualYToDrawingY(tempY).intValue());
                        }
                    }
                }

                if (mCurrentPercentToDraw >= 1f) {
                    mFinishedDrawing = true;
                }
                else if ((mAnimationElapsedTime/mAnimationTime) > 1f) {
                    mCurrentPercentToDraw = 1f;
                    postDelayed(mRedrawRunnable, CoreSettings.ANIMATION_RUNNABLE_RESOLUTION);
                }
                else {
                    mCurrentPercentToDraw = (mAnimationElapsedTime/mAnimationTime);
                    postDelayed(mRedrawRunnable, CoreSettings.ANIMATION_RUNNABLE_RESOLUTION);
                }
            }

            // Draw a line at Y=0 if there is a mix of negative and positive values
            if (getLargestBarValueChartable() > 0f && getSmallestBarValueChartable() < 0f) {
                mDrawingPaint.setColor(mZeroLineColor);
                mDrawingPaint.setStrokeWidth(mZeroLineWidth);
                mDrawingPaint.setStyle(Paint.Style.STROKE);
                mDrawingPaint.setAntiAlias(true);
                pCanvas.drawLine(0f, actualYToDrawingY(0f), mRectF.width(),
                        actualYToDrawingY(0f), mDrawingPaint);
            }

            // Draw the border
            mDrawingPaint.setColor(mBorderColor);
            mDrawingPaint.setStrokeWidth(mBorderWidth);
            mDrawingPaint.setStyle(Paint.Style.STROKE);
            mDrawingPaint.setAntiAlias(true);
            pCanvas.drawRect(mRectF.left, mRectF.top, mRectF.right, mRectF.bottom, mDrawingPaint);
        } else if (isInEditMode()) {
            //The default view which is used, for example, in the layout preview
            mDrawingRect.set(mPadding, mPadding, getWidth() - mPadding, getHeight() - mPadding);
            mDrawingPaint.setColor(Color.BLACK);
            mDrawingPaint.setStrokeWidth(2.0f);
            mDrawingPaint.setStyle(Paint.Style.STROKE);
            mDrawingPaint.setAntiAlias(true);
            pCanvas.drawRect(mDrawingRect.left, mDrawingRect.top, mDrawingRect.right,
                    mDrawingRect.bottom, mDrawingPaint);

            float spacing = getHeight() / 5;
            for (int i = 1; i <= 4; i++) {
                mDrawingPaint.setColor(Color.GRAY);
                mDrawingPaint.setStrokeWidth(1.0f);
                mDrawingPaint.setStyle(Paint.Style.STROKE);
                mDrawingPaint.setAntiAlias(true);
                pCanvas.drawLine(0f, spacing * i, getWidth(), spacing * i, mDrawingPaint);
            }
        }
    }

    protected BarChartItem tappedGetBarChartItem(float pX, float pY) {
        BarChartItem item = null;
        for (int i = 0; i < mTouchRegions.size(); i++) {
            if (mTouchRegions.get(i).contains((int)pX, (int)pY))
                item = mBars.get(i);
        }

        return item;
    }

    protected void reanimate() {
        mCurrentPercentToDraw = 0f;
        mAnimationElapsedTime = 0;
        mAnimationStartTime = 0;
        mFinishedDrawing = false;
        mAnimationCounterStarted = false;
        invalidate();
    }
}
